@@ -174,6 +174,53 @@ static int deriveKeyPair(const uint8_t *seed, const size_t seed_len, uint8_t skS
   return 0;
 }
 
+static int finalize(const uint8_t *x, const uint16_t x_len,
+                         const uint8_t N[crypto_core_ristretto255_BYTES],
+                         uint8_t rwdU[OPRF_BYTES]) {
+
+  // - concat(y, Harden(y, params))
+  uint8_t concated[2*crypto_hash_sha512_BYTES];
+  uint8_t *y=concated, *hardened=concated+crypto_hash_sha512_BYTES;
+  if(-1==sodium_mlock(&concated,sizeof concated)) {
+    return -1;
+  }
+
+  if(0!=oprf_Finalize(x, x_len, N, y)) return 1;
+
+#if (defined TRACE || defined CFRG_TEST_VEC)
+  dump((uint8_t*) y, crypto_hash_sha512_BYTES, "oprf ");
+#endif
+
+#ifdef CFRG_TEST_VEC
+  // testvectors use identity as MHF
+  memcpy(hardened, y, crypto_hash_sha512_BYTES);
+#else
+  // salt - according to the irtf draft this could be all zeroes
+  // TODO parametrize via params and fall back to default
+  uint8_t salt[crypto_pwhash_SALTBYTES]={0};
+  if (crypto_pwhash(hardened, crypto_hash_sha512_BYTES,
+                    (const char*) y, crypto_hash_sha512_BYTES, salt,
+                    crypto_pwhash_OPSLIMIT_INTERACTIVE,
+                    crypto_pwhash_MEMLIMIT_INTERACTIVE,
+                    crypto_pwhash_ALG_DEFAULT) != 0) {
+    /* out of memory */
+    sodium_munlock(concated, sizeof(concated));
+    return -1;
+  }
+#endif
+#if (defined TRACE|| defined CFRG_TEST_VEC)
+  dump(concated, sizeof concated, "concated");
+#endif
+
+  crypto_kdf_hkdf_sha512_extract(rwdU, NULL, 0, concated, sizeof concated);
+  sodium_munlock(concated, sizeof(concated));
+
+#if (defined TRACE|| defined CFRG_TEST_VEC)
+  dump((uint8_t*) rwdU, OPRF_BYTES, "rwdU");
+#endif
+  return 0;
+}
+
 static int prf(const uint8_t *pwdU, const uint16_t pwdU_len,
                const uint8_t kU[crypto_core_ristretto255_SCALARBYTES],
                uint8_t rwdU[OPAQUE_RWDU_BYTES]) {
@@ -205,7 +252,7 @@ static int prf(const uint8_t *pwdU, const uint16_t pwdU_len,
 #endif
 
   // 2. rwdU = Finalize(pwdU, N, "OPAQUE01")
-  if(0!=oprf_Finalize(pwdU, pwdU_len, N, rwdU)) {
+  if(0!=finalize(pwdU, pwdU_len, N, rwdU)) {
     sodium_munlock(N,sizeof N);
     return -1;
   }
@@ -697,24 +744,24 @@ int opaque_Register(const uint8_t *pwdU, const uint16_t pwdU_len,
 //(UsrSession, sid , ssid , S, pw): U picks r, x_u ←_R Z_q ; sets α := (H^0(pw))^r and
 //X_u := g^x_u ; sends α and X_u to S.
 // more or less corresponds to CreateCredentialRequest in the irtf draft
-int opaque_CreateCredentialRequest(const uint8_t *pwdU, const uint16_t pwdU_len, uint8_t _sec[OPAQUE_USER_SESSION_SECRET_LEN+pwdU_len], uint8_t _pub[OPAQUE_USER_SESSION_PUBLIC_LEN]) {
+int opaque_CreateCredentialRequest(const uint8_t *pwdU, const uint16_t pwdU_len, uint8_t _sec[OPAQUE_USER_SESSION_SECRET_LEN+pwdU_len], uint8_t ke1[OPAQUE_USER_SESSION_PUBLIC_LEN]) {
   Opaque_UserSession_Secret *sec = (Opaque_UserSession_Secret*) _sec;
-  Opaque_UserSession *pub = (Opaque_UserSession*) _pub;
+  Opaque_UserSession *pub = (Opaque_UserSession*) ke1;
 #ifdef TRACE
   memset(_sec, 0, OPAQUE_USER_SESSION_SECRET_LEN+pwdU_len);
-  memset(_pub, 0, OPAQUE_USER_SESSION_PUBLIC_LEN);
+  memset(ke1, 0, OPAQUE_USER_SESSION_PUBLIC_LEN);
 #endif
 
   // 1. (blind, blinded) = Blind(pwdU)
   if(0!=oprf_Blind(pwdU, pwdU_len, sec->blind, pub->blinded)) return -1;
 #ifdef TRACE
   dump(_sec,OPAQUE_USER_SESSION_SECRET_LEN+pwdU_len, "sec ");
-  dump(_pub,OPAQUE_USER_SESSION_PUBLIC_LEN, "pub ");
+  dump(ke1,OPAQUE_USER_SESSION_PUBLIC_LEN, "pub ");
 #endif
   memcpy(sec->blinded, pub->blinded, crypto_core_ristretto255_BYTES);
 
   // x_u ←_R Z_q
-#ifndef CFRG_TEST_VEC
+#ifndef CFRG_TEST_VEC // if it is defined then so are the values in this block
   const size_t client_keyshare_seed_len=32;
   uint8_t client_keyshare_seed[client_keyshare_seed_len];
   randombytes(client_keyshare_seed, crypto_scalarmult_SCALARBYTES);
@@ -728,18 +775,17 @@ int opaque_CreateCredentialRequest(const uint8_t *pwdU, const uint16_t pwdU_len,
 #endif
   memcpy(pub->nonceU, sec->nonceU, OPAQUE_NONCE_BYTES);
 
-
   deriveKeyPair(client_keyshare_seed, client_keyshare_seed_len, sec->x_u, pub->X_u);
 
   sec->pwdU_len = pwdU_len;
   memcpy(sec->pwdU, pwdU, pwdU_len);
 
   // keep ke1 for later
-  memcpy(sec->ke1, _pub, OPAQUE_USER_SESSION_PUBLIC_LEN);
+  memcpy(sec->ke1, ke1, OPAQUE_USER_SESSION_PUBLIC_LEN);
 
 #ifdef TRACE
   dump(_sec,OPAQUE_USER_SESSION_SECRET_LEN+pwdU_len, "sec ");
-  dump(_pub,OPAQUE_USER_SESSION_PUBLIC_LEN, "pub ");
+  dump(ke1,OPAQUE_USER_SESSION_PUBLIC_LEN, "pub ");
 #endif
   return 0;
 }
@@ -752,14 +798,14 @@ int opaque_CreateCredentialRequest(const uint8_t *pwdU, const uint16_t pwdU_len,
 // (d) Computes K := KE(p_s, x_s, P_u, X_u) and SK := f K (0);
 // (e) Sends β, X s and c to U;
 // (f) Outputs (sid , ssid , SK).
-int opaque_CreateCredentialResponse(const uint8_t _pub[OPAQUE_USER_SESSION_PUBLIC_LEN], const uint8_t _rec[OPAQUE_USER_RECORD_LEN], const Opaque_Ids *ids, const uint8_t *ctx, const uint16_t ctx_len, uint8_t _resp[OPAQUE_SERVER_SESSION_LEN], uint8_t sk[OPAQUE_SHARED_SECRETBYTES], uint8_t authU[crypto_auth_hmacsha512_BYTES]) {
+int opaque_CreateCredentialResponse(const uint8_t ke1[OPAQUE_USER_SESSION_PUBLIC_LEN], const uint8_t _rec[OPAQUE_USER_RECORD_LEN], const Opaque_Ids *ids, const uint8_t *ctx, const uint16_t ctx_len, uint8_t ke2[OPAQUE_SERVER_SESSION_LEN], uint8_t sk[OPAQUE_SHARED_SECRETBYTES], uint8_t authU[crypto_auth_hmacsha512_BYTES]) {
 
-  Opaque_UserSession *pub = (Opaque_UserSession *) _pub;
+  Opaque_UserSession *pub = (Opaque_UserSession *) ke1;
   Opaque_UserRecord *rec = (Opaque_UserRecord *) _rec;
-  Opaque_ServerSession *resp = (Opaque_ServerSession *) _resp;
+  Opaque_ServerSession *resp = (Opaque_ServerSession *) ke2;
 
 #ifdef TRACE
-  dump(_pub, sizeof(Opaque_UserSession), "session srv pub ");
+  dump(ke1, sizeof(Opaque_UserSession), "session srv pub ");
   dump(_rec, OPAQUE_USER_RECORD_LEN, "session srv rec ");
 #endif
 
@@ -828,7 +874,7 @@ int opaque_CreateCredentialResponse(const uint8_t _pub[OPAQUE_USER_SESSION_PUBLI
   sodium_munlock(response_pad, sizeof response_pad);
 
 #if (defined TRACE || defined CFRG_TEST_VEC)
-  dump(_resp, sizeof (resp->Z) + crypto_scalarmult_BYTES+sizeof(Opaque_Envelope) + sizeof(masking_info.nonce), "resp(z+mn+mr)" );
+  dump(ke2, sizeof (resp->Z) + crypto_scalarmult_BYTES+sizeof(Opaque_Envelope) + sizeof(masking_info.nonce), "resp(z+mn+mr)" );
 #endif
 
   // this is the ake function Response() as per the irtf cfrg draft
@@ -865,7 +911,7 @@ int opaque_CreateCredentialResponse(const uint8_t _pub[OPAQUE_USER_SESSION_PUBLI
   // mixing in things from the irtf cfrg spec
   char preamble[crypto_hash_sha512_BYTES];
   crypto_hash_sha512_state preamble_state;
-  calc_preamble(preamble, &preamble_state, rec->recU.client_public_key, pkS, _pub, resp, ctx, ctx_len, (Opaque_Ids*) ids);
+  calc_preamble(preamble, &preamble_state, rec->recU.client_public_key, pkS, ke1, resp, ctx, ctx_len, (Opaque_Ids*) ids);
   Opaque_Keys keys;
   if(-1==sodium_mlock(&keys,sizeof(keys))) {
     sodium_munlock(x_s,sizeof x_s);
@@ -925,7 +971,7 @@ int opaque_CreateCredentialResponse(const uint8_t _pub[OPAQUE_USER_SESSION_PUBLI
 #ifdef TRACE
   dump(resp->auth, sizeof(resp->auth), "session srv auth ");
   dump(authU, crypto_auth_hmacsha512_BYTES, "authU");
-  dump(_resp, OPAQUE_SERVER_SESSION_LEN, "resp");
+  dump(ke2, OPAQUE_SERVER_SESSION_LEN, "resp");
 #endif
 
   return 0;
@@ -939,21 +985,21 @@ int opaque_CreateCredentialResponse(const uint8_t _pub[OPAQUE_USER_SESSION_PUBLI
 //     Otherwise sets (p_u, P_u, P_s ) := AuthDec_rw (c);
 // (d) Computes K := KE(p_u, x_u, P_s, X_s) and SK := f_K(0);
 // (e) Outputs (sid, ssid, SK).
-int opaque_RecoverCredentials(const uint8_t _resp[OPAQUE_SERVER_SESSION_LEN],
+int opaque_RecoverCredentials(const uint8_t ke2[OPAQUE_SERVER_SESSION_LEN],
                               const uint8_t *_sec/*[OPAQUE_USER_SESSION_SECRET_LEN+pwdU_len]*/,
                               const uint8_t *ctx, const uint16_t ctx_len,
                               const Opaque_Ids *ids0,
                               uint8_t sk[OPAQUE_SHARED_SECRETBYTES],
-                              uint8_t authU[crypto_auth_hmacsha512_BYTES],
+                              uint8_t authU[crypto_auth_hmacsha512_BYTES], // aka ke3
                               uint8_t export_key[crypto_hash_sha512_BYTES]) {
 
-  Opaque_ServerSession *resp = (Opaque_ServerSession *) _resp;
+  Opaque_ServerSession *resp = (Opaque_ServerSession *) ke2;
   Opaque_UserSession_Secret *sec = (Opaque_UserSession_Secret *) _sec;
 
 #ifdef TRACE
   dump(sec->pwdU,sec->pwdU_len, "session user finish pwdU ");
   dump(_sec,OPAQUE_USER_SESSION_SECRET_LEN, "session user finish sec ");
-  dump(_resp,OPAQUE_SERVER_SESSION_LEN, "session user finish resp ");
+  dump(ke2,OPAQUE_SERVER_SESSION_LEN, "session user finish resp ");
 #endif
 
   // 1. (client_private_key, server_public_key, export_key) =
@@ -979,7 +1025,7 @@ int opaque_RecoverCredentials(const uint8_t _resp[OPAQUE_SERVER_SESSION_LEN],
     return -1;
   }
   // 1.2. y = Finalize(pwdU, N, "OPAQUE01")
-  if(0!=oprf_Finalize(sec->pwdU, sec->pwdU_len, N, rwdU)) {
+  if(0!=finalize(sec->pwdU, sec->pwdU_len, N, rwdU)) {
     sodium_munlock(N, sizeof N);
     sodium_munlock(rwdU,sizeof rwdU);
     return -1;
@@ -1247,7 +1293,12 @@ int opaque_CreateRegistrationRequest(const uint8_t *pwdU, const uint16_t pwdU_le
 // (3) computes: β := α^k_s,
 // (4) finally generates: p_s ←_R Z_q, P_s := g^p_s;
 // called CreateRegistrationResponse in the irtf cfrg rfc draft
-int opaque_CreateRegistrationResponse(const uint8_t blinded[crypto_core_ristretto255_BYTES], const uint8_t skS[crypto_scalarmult_SCALARBYTES], uint8_t _sec[OPAQUE_REGISTER_SECRET_LEN], uint8_t _pub[OPAQUE_REGISTER_PUBLIC_LEN]) {
+int opaque_CreateRegistrationResponse_extKeygen(const uint8_t blinded[crypto_core_ristretto255_BYTES],
+                                                const uint8_t skS[crypto_scalarmult_SCALARBYTES],
+                                                uint8_t _sec[OPAQUE_REGISTER_SECRET_LEN],
+                                                uint8_t _pub[OPAQUE_REGISTER_PUBLIC_LEN],
+                                                const toprf_keygencb keygen,
+                                                const void* keygen_ctx) {
   Opaque_RegisterSrvSec *sec = (Opaque_RegisterSrvSec *) _sec;
   Opaque_RegisterSrvPub *pub = (Opaque_RegisterSrvPub *) _pub;
 
@@ -1256,7 +1307,13 @@ int opaque_CreateRegistrationResponse(const uint8_t blinded[crypto_core_ristrett
 
   // k_s ←_R Z_q
   // 1. (kU, _) = KeyGen()
-  oprf_KeyGen(sec->kU);
+  if(keygen==NULL) {
+    oprf_KeyGen(sec->kU);
+  } else {
+    if(0!=keygen(keygen_ctx, sec->kU)) {
+      return -1;
+    }
+  }
 
   // computes β := α^k_s
   // 2. Z = Evaluate(kU, request.data)
@@ -1287,6 +1344,13 @@ int opaque_CreateRegistrationResponse(const uint8_t blinded[crypto_core_ristrett
   return 0;
 }
 
+int opaque_CreateRegistrationResponse(const uint8_t blinded[crypto_core_ristretto255_BYTES],
+                                      const uint8_t skS[crypto_scalarmult_SCALARBYTES],
+                                      uint8_t _sec[OPAQUE_REGISTER_SECRET_LEN],
+                                      uint8_t _pub[OPAQUE_REGISTER_PUBLIC_LEN]) {
+  return opaque_CreateRegistrationResponse_extKeygen(blinded, skS, _sec, _pub, NULL, NULL);
+}
+
 // user computes:
 // (a) Checks that β ∈ G ∗ . If not, outputs (abort, sid , ssid ) and halts;
 // (b) Computes rw := H(key, pw | β^1/r );
@@ -1321,7 +1385,7 @@ int opaque_FinalizeRequest(const uint8_t *_sec/*[OPAQUE_REGISTER_USER_SEC_LEN+pw
     return -1;
   }
   // 2. y = Finalize(pwdU, N, "OPAQUE01")
-  if(0!=oprf_Finalize(sec->pwdU, sec->pwdU_len, N, rwdU)) {
+  if(0!=finalize(sec->pwdU, sec->pwdU_len, N, rwdU)) {
     sodium_munlock(N, sizeof N);
     sodium_munlock(rwdU, sizeof(rwdU));
     return -1;
